#ifndef FIXED_h
#define FIXED_h
#include "param.h"
#include <gsl/gsl_integration.h>

class FO_H:public Pol<double>{
protected:
	CS_params* p;
	double xb;

	double qT;
	double y;
	double c[2];//[E^y,E^-y]

	double M2;
	double sqrts;
	double mp;
	int n_f;
	double gen;
	double a;

	double mu_F;

	double shat;
	double that;
	double uhat;

	VectorPhi Phi[2];


	inline double shat_(double x){
		return (gsl_pow_2(sqrts)*xb*x);
	}

	inline double that_(double x){
		return (M2 - sqrts*mp*c[1]*x);
	}

	inline double uhat_(double x){
		return (M2 - sqrts*mp*c[0]*xb);
	}

	inline double xb_(double x){
		return (-that/(gsl_pow_2(sqrts)*x-sqrts*mp*c[0]));
	}

	inline double x_min_(){
		return GSL_MAX(0,(sqrts*mp*c[0]-M2)
				/(gsl_pow_2(sqrts)-sqrts*mp*c[1]));
	}

	inline void updateparameter(double x){
		that = that_(x);
		xb=xb_(x);
		shat = shat_(x);
		uhat = uhat_(x);
	}

	inline double g_gg(){
		return N_c*(gsl_pow_4(M2) + gsl_pow_4(shat) + gsl_pow_4(that) + gsl_pow_4(uhat))/(shat*that*uhat);
	}

	inline double g_qg(){
		return C_F*(gsl_pow_2(that) + gsl_pow_2(shat))/(-uhat);
	}

	inline double g_gq(){
		return C_F*(gsl_pow_2(uhat) + gsl_pow_2(shat))/(-that);
	}

	inline double g_qbq(){
		return 2.*gsl_pow_2(C_F)*(gsl_pow_2(that) + gsl_pow_2(uhat))/(shat);
	}

	inline double f(double x){

		double result=0.;

		//------ GQ + QG--------
		result+=Phi[1][e_g]*g_qg()*Phi[0].sumofquarks(n_f);

		result+=Phi[0][e_g]*g_gq()*Phi[1].sumofquarks(n_f);

	#ifndef ONLY_MIXED
		//------ GG--------
		result+=Phi[0][e_g]*g_gg()*Phi[1][e_g];

		//------ QQ--------
		double tmp=0.;
		for(e_parton q=e_(n_f); q>e_g; --q){
			tmp+=Phi[0][q]*Phi[1][bar(q)];
			tmp+=Phi[0][bar(q)]*Phi[1][q];
		}
		result+=g_qbq()*tmp;
	#endif

		return result;
	}

	inline static double integrand(double x, void* parameter){
		FO_H* FO = (FO_H*)parameter;
		FO->updateparameter(x);
		FO->Phi[0].update(x,FO->mu_F);
		FO->Phi[1].update(FO->xb,FO->mu_F);
		return (-FO->f(x)/(x*FO->that));
	}

	inline double FO_Z_a(){
		double result, error;
		int status;


		double xmin=x_min_();//cout << " xmin=" << xmin << endl;
		if(xmin<M2*c[0]/(sqrts*mp)) cout << "-- Anderes xmin" << endl;//FIXME
		if(xmin>1.) return 0.;

		int intMem=500000;
		gsl_integration_workspace * w = gsl_integration_workspace_alloc (intMem);
		gsl_function k;
		k.function = &integrand;
		k.params = this;


		//FIXMEstatus = gsl_integration_qags (&k, xmin, 1, gen/100., gen/100., intMem, w, &result, &error);
		status = gsl_integration_qag (&k, xmin, 1.0,  gen, gen, intMem, 3, w, &result, &error);


		if(status!=GSL_SUCCESS){
			cout << "----------- x Integral\t="<< result << " +- " << error << "\t D= "<< error/result*100 << "%" << endl
					<< "                      \tqT = " << p->qT << "\tmu = " << p->mu << endl;
			GSL_ERROR_CODE(status);
		}

		gsl_integration_workspace_free (w);

		return 2.*result/M2;
	}


public:

	inline void init(){
		mp = sqrt(M2+gsl_pow_2(p->qT));
		n_f = p->n_f;
		qT = p->qT;
		mu_F = p->mu;
		a = p->a;

	}

	inline void init(double qT_, double mu_){
		p->update(qT_,mu_);
		init();
	}

	inline void init(CS_params &p_){
		p=(&p_);
		order(p->order);
		M2 = gsl_pow_2(p->M);
		gen = p->gen;
		sqrts = sqrt(p->s);
		Phi[0] = (VectorPhi(p->sign[0]));
		Phi[1] = (VectorPhi(p->sign[1]));
	}

	inline void update(){
		y = p->y;
		c[0]=exp(p->y);
		c[1]=1./c[0];

		_P[0]=0.0;
		if(_order){
			_P[1]=FO_Z_a();
		}
		setX(a);
	}

	inline void update(double y){
		p->update(y);
		update();
	}

	inline FO_H(CS_params &p_){init(p_);}

};







#endif
